use egg::{Extractor, Pattern, RecExpr, Runner};
use solveq::language::{RULES, EquationLanguage, PlusTimesCostFn, FactorizationCostFn};
use solveq::normal_form::analyze3;
use solveq::parse::parse_equation;

static TEST_EQUATIONS: &[&str] = &[
	"(x + 50) * 10 - 150 - 100",
	"(x - 2) * (x + 2) - 0",
	"x ^ 2 - 4",
	"x ^ 2 - 2 - 0",
	"x ^ 2 - (2 * x + 15)",
	"(x ^ 2 - 2 * x - 15) * (x + 5) - 0",
	"x ^ 3 + 3 * x ^ 2 - 25 * x - 75 - 0",
];

fn main() {
	for eq in TEST_EQUATIONS {
		let start = parse_equation(*eq).unwrap();

	//	println!("{:?}", &start);
	// do transformation to left - right = 0

		let mut runner = Runner::default()
			.with_explanations_enabled()
			.with_expr(&start)
			.run(&*RULES);

		let extractor = Extractor::new(&runner.egraph, FactorizationCostFn);

		let (best_cost, best_expr) = extractor.find_best(runner.roots[0]);

		println!("{}", start);
		println!("{:?} {:?}", best_cost, <RecExpr<EquationLanguage> as AsRef<[EquationLanguage]>>::as_ref(&best_expr));


		println!("");
	}

//	let root = runner.roots[0];
//		let egraph = &runner.egraph;
//	let pattern: Pattern<EquationLanguage> = "(+ (* ?a (* x x)) ?c)".parse().unwrap();
//	let matches = pattern.search(&egraph);



//	println!("{:?}", egraph.classes().count());

	// Analyze
//	analyze3(egraph, runner.roots[0]);

	/*
	for class in egraph.classes() {
		if monic_nonconst_polynomial(egraph, class.id).is_some() {
			let (_, best_expr) = extractor.find_best(class.id);
			println!("Monomial: {}", best_expr);
		}
	}

	println!("{:?}", &matches);
	*/


//	println!("{}", runner.explain_equivalence(&start, &best_expr).get_flat_string());
}


/*
fn power_of_x(egraph: &EGraph, eclass: Id) -> Option<usize> {
	for n in &egraph[eclass].nodes {
		match *n {
			EquationLanguage::Unknown => { return Some(1) },
			EquationLanguage::Mul([a,b]) => {
				let Some(left) = power_of_x(egraph, a) else { continue };
				let Some(right) = power_of_x(egraph, b) else { continue };
				return Some(left + right);
			},
			_ => {}
		}
	}
	None
}

fn monomial(egraph: &EGraph, eclass: Id) -> Option<(usize, Rational)> {
	if let Some(deg) = power_of_x(egraph, eclass) {
		return Some((deg, RATIONAL_ONE.clone()));
	}

	for n in &egraph[eclass].nodes {
		match *n {
			EquationLanguage::Mul([a,b]) => {
				let Some(coeff) = egraph[a].data.clone() else { continue };
				let Some(deg) = power_of_x(egraph, b) else { continue };
				return Some((deg, coeff));
			},
			_ => {}
		}
	}
	None
}


// this is either a power_of_x, or a sum of this and a monomial
fn monic_nonconst_polynomial(egraph: &EGraph, eclass: Id) -> Option<Vec<Rational>> {
	let mut result: Vec<Rational> = Vec::new();

	if let Some(deg) = power_of_x(egraph, eclass) {
		result.resize(deg - 1, RATIONAL_ZERO);
		result.push(RATIONAL_ONE.clone());
		return Some(result);
	}

	for n in &egraph[eclass].nodes {
		match *n {
			EquationLanguage::Add([a,b]) => {
				let Some(mut leading) = monic_nonconst_polynomial(egraph, a)
				else { continue };
				let Some(addon) = monomial(egraph, b)
				else { continue };

				if leading.len() <= addon.0 || leading[addon.0] != RATIONAL_ZERO {
					continue;
				}

				leading[addon.0] = addon.1.clone();
				return Some(leading);
			},
			_ => {},
		}
	}
	None
}

*/


/*
fn analyze(egraph: &EGraph, _id: Id) {
	let mut types: HashMap<Id, SpecialTerm> = HashMap::new();
	let mut todo: VecDeque<Id> = VecDeque::new();
	//	todo.push_back(runner.roots[0]);
	for cls in egraph.classes() {
		todo.push_back(cls.id);
	}

	'todo: while todo.len() > 0 {
		let id = todo.pop_front().unwrap();
		if types.contains_key(&id) {
			continue 'todo;
		}

		if let Some(c) = &egraph[id].data {
			types.insert(id, SpecialTerm::Constant(c.clone()));
			continue 'todo;
		}

		'nodes: for n in &egraph[id].nodes {
			match *n {
				EquationLanguage::Unknown => {
					types.insert(id, SpecialTerm::PowerOfX(1));
					continue 'todo;
				},
				EquationLanguage::Mul([a,b]) => {
					if !types.contains_key(&a) {
						todo.push_back(a);
						todo.push_back(id);
						continue 'nodes;
					}

					if !types.contains_key(&b) {
						todo.push_back(b);
						todo.push_back(id);
						continue 'nodes;
					}

					match (&types[&a], &types[&b]) {
						(SpecialTerm::PowerOfX(dega), SpecialTerm::PowerOfX(degb)) => {
							types.insert(id, SpecialTerm::PowerOfX(*dega + *degb));
						},
						(SpecialTerm::Constant(coeff), SpecialTerm::PowerOfX(deg)) => {
							types.insert(id, SpecialTerm::Monomial(*deg, coeff.clone()));
						},
						_ => { continue 'nodes; },
					}
					continue 'todo;
				},
				EquationLanguage::Add([a,b]) => {
					if !types.contains_key(&a) {
						todo.push_front(a);
						todo.push_back(id);
						continue 'todo;
					}

					if !types.contains_key(&b) {
						todo.push_front(b);
						todo.push_back(id);
						continue 'todo;
					}

					match (&types[&a], &types[&b]) {
						(SpecialTerm::MonicNonconstPoly(poly), SpecialTerm::Monomial(deg, coeff)) => {
							if poly.len() <= *deg || poly[*deg] != RATIONAL_ZERO {
								continue 'nodes;
							}

							let mut poly = poly.clone();
							poly[*deg] = coeff.clone();
							types.insert(id, SpecialTerm::MonicNonconstPoly(poly));
						},
						_ => { continue 'nodes; },
					}
					continue 'todo;
				},
				_ => {},
			}
		}

		types.insert(id, SpecialTerm::Other);
	}

	for (id, ty) in &types {
		if !matches!(ty, &SpecialTerm::Other) {
			println!("{:?}", &ty);
		}
	}
}
*/
